
#define CONFIG "/etc/woap/woap.ini"
#define VERSION "1.0-c"
#define PROTOCOL "1.5"

#ifdef WOAP_TELNET
#include <sys/types.h>
#include <sys/stat.h>
#include <termios.h>
#include <stdlib.h>
#include <stdio.h>

#else
#include "includes.h"
#include "ssh.h"
#include "xmalloc.h"
//#include "log.h"

#include<sys/types.h>
#include<unistd.h>
#include<pwd.h>
#endif

#define debug(...) ;

#include<string.h>
#include<stdlib.h>
#include "woap.h"

char* woap_login = NULL;

#define die(error) { fprintf(stderr,error); exit(1); }

#ifndef WOAP_TELNET
char* woap_user(const char* client_user) {
	struct passwd *pwd;
	pwd = getpwuid(getuid());
	woap_login = xstrdup(client_user);
	debug("woap_user('%s')",client_user);
	return pwd ? xstrdup(pwd->pw_name) : NULL;
}

int woap_auth(const char* client_password) {
	char* ini_password;
	if(!woap_login || !client_password) exit(1);
	ini_password = woap_config("password",woap_login,NULL);

	debug("woap_auth(%s/%s) ini_password: %s\n",woap_login,
						client_password,ini_password);

	if(!ini_password) {
		fprintf(stderr,"Rejected user %s: not in list of allowed users\n",
						woap_login);
		return 0;
	}
	if(strcmp(ini_password,client_password)) {
		fprintf(stderr,"%s: Password mismatch\n",woap_login);
		return 0;
	}
	return 1;
}
#endif

#include<sys/utsname.h>
#include<unistd.h>
#include<string.h>
#include<time.h>
#include<ctype.h>

#define BUF_MAX 16384

char* g_session_id = "C2BF612A-F2F9-4F6A-B82C-A3B2451BE5D8";


int get_ip_32bit(char* ip) {
	char *a=ip,*b=NULL,*c=NULL,*d=NULL;
	if(a && strchr(a,'.')) b = strchr(a,'.')+1;
	if(b && strchr(b,'.')) c = strchr(b,'.')+1;
	if(c && strchr(c,'.')) d = strchr(c,'.')+1;
	if(!d) return 0;
	return atoi(a)*16777216+atoi(b)*65536+atoi(c)*256+atoi(d);	
}

char* random_hex(char* where) {
	char hex[] = "0123456789ABCDEF";
	*where++ = hex[random()%16]; *where++ = hex[random()%16];
	*where++ = hex[random()%16]; *where++ = hex[random()%16];
	return where;
}

char* generate_session_id() {
	char buff[BUF_MAX], *p = buff;
	srandom(time(NULL));
	p = random_hex(p); p = random_hex(p); *p++ = '-';
	p = random_hex(p); *p++ = '-';
	p = random_hex(p); *p++ = '-';
	p = random_hex(p); *p++ = '-';
	p = random_hex(p); p = random_hex(p); p = random_hex(p);
	*p = '\0';
	return g_session_id = strdup(buff);
}

char* chomp(char* line) {
	if(line[strlen(line)-1] == '\n') line[strlen(line)-1] = '\0';
	if(line[strlen(line)-1] == '\r') line[strlen(line)-1] = '\0';
	if(line[strlen(line)-1] == '\n') line[strlen(line)-1] = '\0';
	return line;
}
char* woap_config(char* section,char* key,char* default_value) {
	char current[BUF_MAX] = ""; /* current session name */
	char buff[BUF_MAX];
	FILE *fd;

	if(!(fd = fopen(CONFIG,"r"))) 
		return default_value ? strdup(default_value) : default_value;

	while(fgets(buff,BUF_MAX,fd)) {
		chomp(buff);

		if(buff[0] == ';' || buff[0] == '#' || 
		   buff[0] == 0 || buff[0] == '\n') continue;
		if(buff[0] == '[' && buff[strlen(buff)-1] == ']') {
			buff[strlen(buff)-1] = '\0';
			strcpy(current,buff+1);
			continue;
		}
		debug("woap: config: [%s] %s\n",current,buff);
		if(!strcmp(current,section) &&
		   !strncmp(key,buff,strlen(key)) && buff[strlen(key)] == '=') {
			fclose(fd);
			return strdup(&buff[strlen(key)+1]);
		}
	}
	fclose(fd);
	return default_value ? strdup(default_value) : default_value;
}
int parse_nums(char* line,int num[],int num_max) {
	int nums = 0;
	// Mem: 7983160K av, 7975524K used, 7636K free, 409100K shrd, 87760K buff
	//  -> [ 7983160, 7975524, 7636, 409100, 87760 ]
	while(1) {
		while(*line && !isdigit(*line)) line++;
		if(!*line) break;
		num[nums++] = strtol(line,&line,10);
	}
	return nums;
}

void array_free(char* array[]) {
	int i;
	for(i=0;array[i];i++) free(array[i]);
}

int array_add(char* array[],int *pos,int max,char* key,int number,char* value) {
	char buff[BUF_MAX];
	if((*pos)+2 >= max) exit(10);

	if(number) sprintf(buff,"%s%d",key,number);
		  else strcpy(buff,key);
	array[(*pos)+0] = strdup(buff);
	array[(*pos)+1] = strdup(value);
	array[(*pos)+2] = NULL; 
	*pos = (*pos) + 2;
	return (*pos);
}

char* woap_pack(char* param[]) {
	char output[BUF_MAX] = "#";
	int i;
	for(i=0;param[i];i+=2) {
		strcat(output,param[i+0]); 
		strcat(output,"=");
		if(param[i+1]) strcat(output,param[i+1]); 
		strcat(output,"#");
	}
	return strdup(output);
}

int woap_unpack(char* buffer,char* param[],int param_max) {
	char* copy = strdup(buffer); char *input = copy;
	char *value_quote,*value_end;

	int i=0;
	if(*input++ != '#') exit(8);
	while(*input) {
		char* value = strchr(input,'=');
		if(!value) return 0;
		*value++ = '\0';
		value_quote = strchr(value+1,'"');
		value_end   = strchr(value,'#');

		// s/^\"([^\"]*)\"\#// or s/^([^\#]*)\#//
		if(value[0] == '\"' && value_quote[1] == '#')
			value++, value_end = value_quote+1, value_quote[0] = '\0'; 
		value_end[0] = '\0'; 

		array_add(param,&i,param_max,input,0,value);
		input = value_end+1;
	}
	param[i] = NULL, param[i+1] = NULL;
	free(copy);
	return i;
}

int send_cmd(char* cmd,char* param[]) {
	char* packet;
	char* msg[BUF_MAX] = { "keycount", "", "responseaction", NULL,
						  "sessionid", NULL };
	int i,len;

	msg[3] = cmd; msg[5] = g_session_id;

	for(i=0;param[i];i++) msg[6+i] = param[i], msg[6+i+1] = NULL;
	msg[1] = strdup("SPACE");
	snprintf(msg[1],strlen(msg[1]),"%d",1+1+i/2);

	packet = woap_pack(msg);
	len = write(fileno(stdout),packet,strlen(packet));
	fflush(stdout);
	fprintf(stderr,"> %s\n",packet);
	if(len != (int)strlen(packet)) die("send_cmd: failed to write packet\n");
	free(packet);
	return len;
}

char* read_cmd(char* args[],int args_max) {
	char buffer[BUF_MAX] = "";
	char *cmd,*param[BUF_MAX],*sessionid;
	int  i,len,received_kc,keycount;

	while(1) {
		len = read(fileno(stdin),buffer+strlen(buffer),4096);
		fprintf(stderr,"< [%d] %s\n",len,buffer);
		if(len <= 0) return NULL; // todo: if <0, die with error

		received_kc = woap_unpack(buffer,param,BUF_MAX);
		if(!received_kc) continue;

		if(strcmp(param[0],"keycount")) 	 die("no keycount\n");
		keycount = atoi(param[1]);
		if(keycount) ; // we should check keycount..
		if(strcmp(param[2],"requestaction")) die("no requestaction\n");
		cmd = strdup(param[3]);
		if(strcmp(param[4],"sessionid")) 	 die("no sessionid\n");
		sessionid = param[5];
		if(strcmp(sessionid,g_session_id))   die("Session ID mismatch\n");

		for(i=0;param[i];i+=2) {
			if(i-6 >= args_max) die("should not happen");
			if(i>=6) args[i-6] = param[i], args[i-6+1] = param[i+1];
			// warn("%d %s=%s\n",i,param[i],param[i+1]);
			if(i<6)  free(param[i]), free(param[i+1]);
		}
		args[i-6] = NULL;
		return cmd;
	}
}

int get_disks(char* disks[],int disks_max,char* free) {
	char *percent,*mountpoint,*cmd = "/bin/df",buff[BUF_MAX];
	int capacity=0,i=0;
	struct stat foo;
	FILE *df;
	if(!stat("/usr/xpg4/bin/df",&foo)) cmd = "/usr/xpg4/bin/df -P";
	if(!(df = popen(cmd,"r"))) return 0;
	while(fgets(buff,BUF_MAX,df)) {
		chomp(buff);
		percent = strchr(buff,'%');
		if(!percent || !isdigit(percent[-1])) continue;
		mountpoint = percent+1;
		while(*mountpoint == ' ') mountpoint++;
		*percent-- = '\0';
		while(isdigit(*percent)) percent--;
		if(free && !strcmp(free,mountpoint)) capacity = atoi(percent);
		disks[i++] = strdup(mountpoint);
	}
	pclose(df);
	disks[i] = NULL;
	return capacity;
}

int get_processes(char* commands[],int commands_max,char* running) {
	struct utsname machine;
	char *slash,*cmd,buff[BUF_MAX],*user,*command,*p;
	int i=0,status = 0;
	FILE *ps;

	if(uname(&machine)) exit(2);
	if(!strcmp(machine.sysname,"IRIX64") || 
	   !strcmp(machine.sysname,"SunOS")) cmd = "ps -e -o user,args";
									else cmd = "ps -ax -o user,args";
	
	if(!(ps = popen(cmd,"r"))) return 0;
	fgets(buff,BUF_MAX,ps); // read header line
	while(fgets(buff,BUF_MAX,ps)) {
		chomp(buff);
		user = buff;
		while(*user == ' ') user++;
		command = user;
		while(*command != ' ') command++;
		*command++ = '\0';
		while(*command == ' ') command++;
		if(strcmp(user,"root") && strcmp(user,"woap")) continue;
		fprintf(stderr,"user: %s, command: <%s>\n",user,command);

		p = command, slash = p;
		if(*command != '[')
			while(*p && *p != ' ') if(*p++ == '/') slash = p;

		while(command[strlen(command)-1] == ' ') 
				command[strlen(command)-1] = '\0';
		commands[i++] = strdup(slash);

		if(running && !strncmp(running,slash,strlen(running)))
				status++;
	}
	commands[i] = NULL;
	pclose(ps);
	return status;
}

#define TYPE_NONE   0
#define TYPE_MEM    1
#define TYPE_SWAP   2
#define TYPE_MEMORY 3

void get_free_save(int *where,int used,int total) {
	if(where) *where = total ? 100-(100*used/total) : 0;
}

int get_free(int *ram,int *swap) {
	char buff[BUF_MAX];
	int type,nums,num[BUF_MAX];
	int _used,_free;
	FILE *top;

	if(!(top = popen("top -d 1 -n 1 -b","r"))) return 0;
	while(fgets(buff,BUF_MAX,top)) {
		chomp(buff);
		nums = parse_nums(buff,num,BUF_MAX);

		type = TYPE_NONE; 
		if(!strncmp(buff,"Mem:",4))    type = TYPE_MEM;
		if(!strncmp(buff,"Swap:",5))   type = TYPE_SWAP;
		if(!strncmp(buff,"Memory:",7)) type = TYPE_MEMORY;

        // Memory: Real: 16M/47M act/tot  Free: 198M  Swap: 0K/128M used/tot
        if(!strncmp(buff,"Memory: Real:",13)) {
			_used = num[1]; _free = num[2];
			get_free_save(ram, _used,_used+_free);
			get_free_save(swap,num[3],num[4]);
			continue;
		}

		if(type == TYPE_NONE) continue;
		//printf("get_free %s\n",buff);

		if(type == TYPE_MEM && (nums == 4 || nums == 5))
			get_free_save(ram,num[1],num[0]);
		if(type == TYPE_MEM && nums == 6) {
			_used = num[0] + num[2] + num[4];
			_free = num[1] + num[3] + num[5];
			get_free_save(ram,_used,_used+_free);
		}
		if(type == TYPE_SWAP && nums == 4)
			get_free_save(swap,num[2],num[0]);
		if(type == TYPE_SWAP && nums == 2)
			get_free_save(swap,num[0]-num[2],num[0]);
		if(type == TYPE_MEMORY && nums == 5) {
			get_free_save(ram,num[0]-num[2],num[0]);
			get_free_save(swap,num[3]-num[4],num[3]);
		}
		if(type == TYPE_MEMORY && nums == 4) {
			get_free_save(ram,num[0]-num[1],num[0]);
			get_free_save(swap,num[2]-num[3],num[2]);
		}
	}
	pclose(top);
	return 1;
}

void free_check(char* arg[],int *status,char* result,int check_ram,
				 char* var,char* msg_ok,char* msg_fail) {
	int ram=-1,swap=-1,free,limit;

	if(strcmp(arg[0],var)) return;
	limit = atoi(arg[1]);
	get_free(&ram,&swap);
	free = check_ram ? ram : swap;
	*status = (free > limit);
	sprintf(result,*status ? msg_ok : msg_fail,free);
}

int param_combo(char* param[],int *pos,int param_max,char* name,int number,
				char* values[]) {
	char buf[BUF_MAX];
	int i;
	array_add(param,pos,param_max,"paramname",number,name);
	array_add(param,pos,param_max,"paramtype",number,"listcombo");
	for(i=0;values[i];i++) /* do nothing */;
	sprintf(buf,"%d",i);
	array_add(param,pos,param_max,"valuecount",0,buf);
	for(i=0;values[i];i++) {
		array_add(param,pos,param_max,"value",i+1,values[i]);
	}
	array_add(param,pos,param_max,"atbutton_paramcount",0,"1");
	array_add(param,pos,param_max,"atbutton_param",1,name);
	return i;
}

int params_cpu(char* param[],int param_max) {
	int i = 0;
	array_add(param,&i,param_max,"paramname",1,"Load Limit [%]");
	array_add(param,&i,param_max,"paramtype",1,"longinteger");
	array_add(param,&i,param_max,"maxlength",1,"200");
	return i;
}
int params_diskspace(char* param[],int param_max) {
	char *disks[BUF_MAX] = { NULL };
	int i = 0;
	array_add(param,&i,param_max,"paramname",1,"Disk Free Limit [%]");
	array_add(param,&i,param_max,"paramtype",1,"longinteger");
	array_add(param,&i,param_max,"maxlength",1,"100");

	get_disks(disks,BUF_MAX,NULL);
	param_combo(param,&i,param_max,"Disks",2,disks);
	array_free(disks);
	return i;
}
int params_ram(char* param[],int param_max) {
	int i = 0;
	array_add(param,&i,param_max,"paramname",1,"Free RAM Limit [%]");
	array_add(param,&i,param_max,"paramtype",1,"longinteger");
	array_add(param,&i,param_max,"maxlength",1,"100");
	return i;
}
int params_swap(char* param[],int param_max) {
	int i = 0;
	array_add(param,&i,param_max,"paramname",1,"Free Swap Limit [%]");
	array_add(param,&i,param_max,"paramtype",1,"longinteger");
	array_add(param,&i,param_max,"maxlength",1,"100");
	return i;
}
int params_processes(char* param[],int param_max) {
	char *processes[BUF_MAX] = { NULL };
	int i = 0;
	get_processes(processes,BUF_MAX,NULL);
	param_combo(param,&i,param_max,"Select Process",1,processes);
	array_free(processes);
	return i;
}

void run_cpu(char* arg[],int *status,char result[]) {
	char buff[BUF_MAX],*end;
	int load,limit,nums=0,offset = 3,num[BUF_MAX];
	FILE *cmd;

	if(strcmp(arg[0],"Load Limit [%]")) return;
	limit = atoi(arg[1]);

	if(!(cmd = popen("vmstat 1 2","r"))) return;
	fgets(buff,BUF_MAX,cmd); // header 1
	fgets(buff,BUF_MAX,cmd); // header 2
	chomp(buff);
	end = buff+strlen(buff);
	if(end[-2]=='w' && end[-1]=='a') offset = 4;
	if(end[-5]=='w' && end[-4]=='a') offset = 5;
	fgets(buff,BUF_MAX,cmd); // first line
	fgets(buff,BUF_MAX,cmd); // second line
	chomp(buff);
	pclose(cmd);

	// analyze
	nums = parse_nums(buff,num,BUF_MAX);
	if(nums < 4) return;
	load = num[nums-offset]+num[nums-offset+1];
	*status = (load <= limit);
	if(*status)  sprintf(result,"CPU load is %d",load);
			else sprintf(result,"CPU load %d out of Limit %d",load,limit);
}

void run_diskspace(char* arg[],int *status,char result[]) {
	char *disk,*disks[BUF_MAX] = { NULL };
	int limit,used;

	if(strcmp(arg[0],"Disk Free Limit [%]")) return;
	limit = atoi(arg[1]);
	if(strcmp(arg[2],"Disks")) return;
	disk = arg[3];

	used = get_disks(disks,BUF_MAX,disk);
	*status = (used > limit);
	sprintf(result,*status ? "Disk %s is %d%% used" : 
		"Out of disk space on disk %s: %d%% used",disk,used);
}

void run_ram(char* arg[],int *status,char result[]) {
	free_check(arg,status,result,1,"Free RAM Limit [%]",
			"%d%% of RAM memory free","Low RAM memory: %d%% free");
}
void run_swap(char* arg[],int *status,char result[]) {
	free_check(arg,status,result,0,"Free Swap Limit [%]",
			"%d%% of SWAP memory free","Low SWAP memory: %d%% free");
}
void run_processes(char* arg[],int *status,char result[]) {
	char *process,*processes[BUF_MAX] = { NULL };

	if(strcmp(arg[0],"Select Process")) return;
	process = arg[1];

	*status = get_processes(processes,BUF_MAX,process);

	sprintf(result,*status ? "Process %s is running" :
				"Process %s is not running",process);
}

struct check_t {
	char *name,*desc;
	int (*params)(char *param[],int param_max);
	void (*run_check)(char *param[],int *status,char result[]);
} checks[] = {
	{ "CPU",       "CPU Usage",         params_cpu,       run_cpu },
	{ "DiskSpace", "DiskSpace Usage",   params_diskspace, run_diskspace },
	{ "RAM",       "RAM Usage",         params_ram,       run_ram },
	{ "Swap",      "Swap Usage",        params_swap,      run_swap },
	{ "Processes", "Running processes", params_processes, run_processes },
	{ NULL, NULL }
};

int check_params(char* checkname,char* reply[],int reply_max) {
	char param_string[BUF_MAX];
	char* param[BUF_MAX] = { NULL };
	struct check_t* check;
	int i=0;

	for(check=checks;check->name;check++)
		if(!strcmp(check->name,checkname)) break;
	if(!check->name) return 
		array_add(reply,&i,reply_max,"errormessage",0,"Check not found");

	(check->params)(param,BUF_MAX);
	snprintf(param_string,BUF_MAX,"\"%s\"",woap_pack(param));
	array_free(param);

	array_add(reply,&i,reply_max,"checkid",0,   checkname);
	array_add(reply,&i,reply_max,"parameters",0,param_string);
	return i;
}

int rpc_listchecks(char* cmd,char* param[],char* reply[],int reply_max) {
	struct check_t* check;
	char buff[BUF_MAX];
	int i=0,n=0;

	reply[i++] = strdup("checkcount");
	reply[i++] = NULL; // will be filled later

	for(check=checks;check->name;check++,n++) {
		array_add(reply,&i,reply_max,"checkid",n+1,check->name);
		array_add(reply,&i,reply_max,"checkdescription",n+1,check->desc);
		array_add(reply,&i,reply_max,"available",n+1,"true");
	}
	sprintf(buff,"%d",n);
	reply[1] = strdup(buff);
	return i;
}

int rpc_definecheck(char* cmd,char* param[],char* reply[],int reply_max) {
	if(strcmp(param[0],"checkid")) exit(3);
	return check_params(param[1],reply,reply_max);
}
void rpc_runcheck(char* cmd,char* param[],char* reply[],int reply_max) {
	char resultstring[BUF_MAX] = "No result";
	struct check_t* check;
	char *checkname,*args[BUF_MAX];
	int i=0,status = -1;

	if(strcmp(param[0],"checkid")) exit(4);
	checkname = param[1];
	if(strcmp(param[2],"para")) exit(5);
	woap_unpack(param[3],args,BUF_MAX);

	for(check=checks;check->name;check++)
		if(!strcmp(check->name,checkname)) break;
	if(!check->name) {
		array_add(reply,&i,reply_max,"errormessage",0,"Check not found");
		return;
	}

	(check->run_check)(args,&status,resultstring);
	if(status == -1) {
		array_add(reply,&i,reply_max,"errormessage",0,resultstring);
	} else {
		array_add(reply,&i,reply_max,"checkid",0,checkname);
		array_add(reply,&i,reply_max,"result",0,status ? "up" : "down");
		array_add(reply,&i,reply_max,"resultstring",0,resultstring);
	}
	//if(resultstring) free(resultstring);
}

int rpc_getcombo(char* cmd,char* param[],char* reply[],int reply_max) {
	if(strcmp(param[0],"checkid")) exit(6);
	return check_params(param[1],reply,reply_max);
}

int verify_ipaddress(char* ip,char *login) {
	char *firewall,*min,*max;
	int ip32,min32,max32;

	if(!ip || !login) return 1;
	while(strchr(ip,':')) ip = strchr(ip,':') + 1;
	if(!*ip) return 1; // ?

	firewall = woap_config("firewall",login,NULL);
	if(!firewall) return 1; // no limit for user - it's ok
	min = strdup(firewall);
	max = strchr(min,'-');
	if(!max) return 1; // weird format

	*max++ = '\0'; // split min and max

	ip32 = get_ip_32bit(ip);
	min32 = get_ip_32bit(min);
	max32 = get_ip_32bit(max);

	if(ip32 < min32 || ip32 > max32) {
		fprintf(stderr,"Rejected IP address %s: not in allowed range %s\n",
						ip,firewall);
		exit(1);
	}
	return 1;
}

int woap_agent() {
	char *login[] = {"registrationlevel","standard",NULL};
	char *ip,*ip_end,*cmd;
	char *param[BUF_MAX];
	char *reply[BUF_MAX];

	generate_session_id();

	debug("woap: woap_agent() called...\n");

	ip = getenv("SSH_CLIENT");
	ip_end = ip ? strchr(ip,' ') : NULL;
	if(ip_end) *ip_end = '\0';

	freopen(woap_config("agent","log","/dev/null"),"a",stderr);
	fprintf(stderr,"\n");

	verify_ipaddress(ip,woap_login);

#ifndef WOAP_TELNET
	system("stty raw -echo");
#endif

	fprintf(stderr,"Accepted new connection for user %s from IP %s\n",
			woap_login?woap_login:"<unknown>",ip?ip:"<unknown>");
	fflush(stderr);

	send_cmd("login",login);
	while((cmd = read_cmd(param,BUF_MAX))) {
		reply[0] = NULL;

		if(!strcmp(cmd,"listchecks"))  
				rpc_listchecks(cmd,param,reply,BUF_MAX);
		if(!strcmp(cmd,"definecheck")) 
				rpc_definecheck(cmd,param,reply,BUF_MAX);
		if(!strcmp(cmd,"runcheck"))    
				rpc_runcheck(cmd,param,reply,BUF_MAX);
		if(!strcmp(cmd,"getcomboparamvalues")) 
				rpc_getcombo(cmd,param,reply,BUF_MAX);

		if(!reply[0]) {
			fprintf(stderr,"Command %s not implemented\n",cmd);
			exit(7);
		}
		send_cmd(cmd,reply);
	}
	fflush(stderr);
	return 0;
}

#ifdef WOAP_TELNET
int main(int argc,char* argv[]) {
	//int ram,swap;
	//get_free(&ram,&swap);
	//printf("ram: %d, swap: %d\n",ram,swap);
	//return 0;
	return woap_agent(argc,argv);
}
#endif
